/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getMainCFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getSingleUnitMainCFile;

import java.util.List;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.generator.common.model.source.AbstractUnit;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.Receiver;
import org.fortiss.af3.platform.model.Transmitter;
import org.fortiss.af3.platform.raspberry.model.ActuatorPWM;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button1;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button2;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button3;
import org.fortiss.af3.platform.raspberry.model.gamepad.Button4;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonL1;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonL2;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonR1;
import org.fortiss.af3.platform.raspberry.model.gamepad.ButtonR2;
import org.fortiss.af3.platform.raspberry.model.gamepad.GamepadReceiverBase;
import org.fortiss.af3.platform.raspberry.model.gamepad.Left_StickX_Position;
import org.fortiss.af3.platform.raspberry.model.gamepad.Left_StickY_Position;
import org.fortiss.af3.platform.raspberry.model.gamepad.Right_StickX_Position;
import org.fortiss.af3.platform.raspberry.model.gamepad.Right_StickY_Position;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;

/**
 * Separate class for generating the main.c file.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
class MainGenerator {
	private static final int WAITING_SLEEP_IN_MICROS = 250;
	private RaspberryPi executionUnit;
	private List<Pair<ExecutionUnit, Component>> deployedComponents;
	private List<Pair<PlatformConnectorUnit, Port>> deployedPorts;
	private ITransformationContext context;

	private boolean useGamepad;
	private boolean usePWM;

	/** Constructor. */
	public MainGenerator(RaspberryPi executionUnit,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		this.executionUnit = executionUnit;
		this.deployedComponents = deployedComponents;
		this.deployedPorts = deployedPorts;
		this.context = context;
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(!useGamepad && p.getFirst() instanceof GamepadReceiverBase) {
				useGamepad = true;
			}
			if(!usePWM && p.getFirst() instanceof ActuatorPWM) {
				usePWM = true;
			}
		}
	}

	/** Creates the main.c file for deployments with distributed execution units. */
	public AbstractUnit createMain() {
		String includes = createIncludes(deployedComponents, deployedPorts);
		// TODO: compute syncBox size
		int syncBoxSize = 0;
		// TODO: syncbox code
		String initCode = createInitCode(deployedComponents);
		String workerCode = createWorkerCode(deployedComponents, deployedPorts);
		return getMainCFile(executionUnit.getName(), executionUnit.isCoordinatorUnit(),
				executionUnit.getCanCoordinationID(), executionUnit.getCycleTime(),
				WAITING_SLEEP_IN_MICROS, includes, syncBoxSize, "// TODO\n", initCode, workerCode);
	}

	/** Creates the main.c file for deployments with a single execution units. */
	public AbstractUnit createSingleUnitMain() {
		String includes = createIncludes(deployedComponents, deployedPorts);
		String initCode = createInitCode(deployedComponents);
		String workerCode = createWorkerCode(deployedComponents, deployedPorts);
		return getSingleUnitMainCFile(executionUnit.getName(), executionUnit.getCycleTime(),
				includes, initCode, workerCode);
	}

	/** Creates the includes of the system headers. */
	private String createIncludes(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append("#include <" + c.getName() + "_ID_" + c.getId() + ".h>\n");
		}
		if(useGamepad) {
			sb.append("#include <gamepad.h>\n");
		}
		if(usePWM) {
			sb.append("#include <temp_actuator.h>\n");
		}
		return sb.toString();
	}

	/** Create the worker code. */
	private String createWorkerCode(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Receiver && p.getSecond() instanceof InputPort) {
				sb.append(createReadCode((Receiver)p.getFirst(), (InputPort)p.getSecond()));
			}
		}
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append(makeCall("perform_step", c));
		}
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Transmitter && p.getSecond() instanceof OutputPort) {
				sb.append(createWriteCode((Transmitter)p.getFirst(), (OutputPort)p.getSecond()));
			}
		}
		return sb.toString();
	}

	/** Returns the port and ID identifier. */
	private String portName(Port port) {
		return port.getName() + "_ID_" + port.getId();
	}

	/** Creates the read code for the given receiver and port. */
	private String createWriteCode(Transmitter transmitter, OutputPort outport) {
		if(transmitter instanceof ActuatorPWM) {
			ActuatorPWM act = (ActuatorPWM)transmitter;
			String result = "if(!noval_" + portName(outport) + ") {\n";
			result +=
					"temp_actuator_set_target(" + act.getChannelID() + ", (uint16_t)" +
							portName(outport) + ");\n";
			result += "}\n";
			return result;
		}
		return "";
	}

	/** Creates the read code for the given receiver and port. */
	private String createReadCode(Receiver receiver, InputPort inport) {
		if(receiver instanceof Button1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_1", inport);
		}
		if(receiver instanceof Button2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_2", inport);
		}
		if(receiver instanceof Button3) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_3", inport);
		}
		if(receiver instanceof Button4) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_4", inport);
		}
		if(receiver instanceof ButtonL1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_L1", inport);
		}
		if(receiver instanceof ButtonL2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_L2", inport);
		}
		if(receiver instanceof ButtonR1) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_R1", inport);
		}
		if(receiver instanceof ButtonR2) {
			return gamepadReadCode("gamepad_get_button_state", "GAMEPAD_BUTTON_R2", inport);
		}
		if(receiver instanceof Right_StickX_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_RIGHT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof Right_StickY_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_RIGHT_STICK_VERTICAL", inport);
		}
		if(receiver instanceof Left_StickX_Position) {
			return gamepadReadCode("gamepad_get_axis_position",
					"GAMEPAD_AXIS_LEFT_STICK_HORIZONTAL", inport);
		}
		if(receiver instanceof Left_StickY_Position) {
			return gamepadReadCode("gamepad_get_axis_position", "GAMEPAD_AXIS_LEFT_STICK_VERTICAL",
					inport);
		}
		return "";
	}

	/** Creates read code for gamepad state. */
	private String gamepadReadCode(String gamepadFun, String btnId, InputPort inport) {
		String noval = "noval_" + portName(inport) + " = false;\n";
		String setBtn = portName(inport) + " = " + gamepadFun + "(" + btnId + ");\n";
		return noval + setBtn;
	}

	/** Create the initialize code. */
	private String createInitCode(List<Pair<ExecutionUnit, Component>> deployedComponents) {
		StringBuilder sb = new StringBuilder();
		if(usePWM) {
			sb.append("temp_actuator_initialize(\"/dev/ttyACM0\");\n\n");
		}
		if(useGamepad) {
			sb.append("gamepad_configuration_t* gamepad_config = malloc(sizeof(gamepad_configuration_t));\n");
			sb.append("gamepad_config->device_id = \"/dev/input/js0\";\n");
			sb.append("gamepad_config->waiting_sleep_in_micros = " + WAITING_SLEEP_IN_MICROS +
					";\n");
			sb.append("gamepad_config->axis_callback = NULL;\n");
			sb.append("gamepad_config->button_callback = NULL;\n");
			sb.append("gamepad_initialize(gamepad_config);\n\n");
		}
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append(makeCall("init", c));
		}
		return sb.toString();
	}

	/** Creates the function call code. */
	private String makeCall(String function, Component c) {
		StringBuilder sb = new StringBuilder();
		sb.append(function).append('_').append(c.getName());
		sb.append("_ID_").append(c.getId()).append("();\n");
		return sb.toString();
	}
}
